import math
import random
import cPickle as pickle
from neat import config, population, chromosome, genome, visualize
from neat.nn import nn_pure as nn
import inspect
from pprint import pprint
import numpy as np

from PIL import Image, ImageChops
#from neat.nn import nn_cpp as nn # C++ extension

config.load('CPPN_config')

# set node gene type
chromosome.node_gene_type = genome.NodeGene

# set basic geometry
SUBSTRATE = nn.FeedForward([16,16,1], False, math.tanh)


INPUTS = [[-1.0, 1.0], [-0.33, 1.0], [0.33, 1.0], [1.0, 1.0],  
          [-1.0, 0.33], [-0.33, 0.33], [0.33, 0.33], [1.0, 0.33],  
          [-1.0, -0.33], [-0.33, -0.33], [0.33, -0.33], [1.0, -0.33],
          [-1.0, -1.0], [-0.33, -1.0], [0.33, -1.0], [1.0, -1.0]]

OUTPUTS = [0, 0, 1, 0,   1, 1 , 1, 1,    0, 0, 1, 0,     0, 0, 1, 0]
#OUTPUTS = [0, 0, 0, 0,   1, 1 , 1, 1,    0, 0, 0, 0,     0, 0, 0, 0]
#OUTPUTS = [1, 0, 0, 0,   0, 1 , 0, 0,    0, 0, 1, 0,     0, 0, 0, 1]
OUTPUTS = [0.1, 0.9, 0.9, 0.9,   0.1, 0.9 , 0.1, 0.1,    0.1, 0.1, 0.9, 0.1,     0.9, 0.9, 0.9, 0.1]
#OUTPUTS = [0, 1, 1, 1,   0, 0 , 1, 0,    0, 0, 1, 0,     0, 0, 1, 0]
#OUTPUTS = [0, 0, 0, 1,   0, 0 , 0, 0,    0, 0, 0, 0,     1, 0, 0, 0]


#crude training
SUBSTRATE_INPUT = [[0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1],
                   [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0],
                   [1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]]

#correct classes
SUBSTRATE_OUTPUT = [1,0,1]

correct_im = Image.open("Skjermbilde.JPG")
correct_im = correct_im.convert("L")

pix = np.array(Image.new("L", (200, 140)))

pix.shape = 200,140


correct_pix = correct_im.load()

dataset_min = -1.0
dataset_max = 1.0
output_max = 1.0
output_min = 0.0
dataset_range = dataset_max - dataset_min
output_range = output_max - output_min

def convert_value(val, dataset_min, dataset_range, output_min, output_range):
    return (((val - dataset_min) * output_range) / dataset_range) + output_min


# produce hyper-cube inputs (coordinates x1 y1 x2 y2)
def fully_connected_coord_input(hypercube_dimensions = 4):
    hc_range = range(hypercube_dimensions)
    return [[x1, y1, x2, y2] for x1 in hc_range for y1 in hc_range for x2 in hc_range for y2 in hc_range [x1, y1, x2, y2]]


#H_INPUTS = fully_connected_coord_input()        

def eval_fitness(population):
    
    for chromo in population:
        net = nn.create_ffphenotype(chromo)

        error = 0.0
        for i, inputs in enumerate(INPUTS):
            net.flush() # not strictly necessary in feedforward nets
            output = net.sactivate(inputs) # serial activation
            error += (output[0] - OUTPUTS[i])**2

        chromo.fitness = 1.0/(1.0 + error)
        
        
def eval_image_fitness(population):
    for chromo in population:
        net = nn.create_ffphenotype(chromo)
        
        for x in range(200):
            for y in range(140):
                intensity = net.sactivate([x, y, math.sqrt(math.fabs(100-x)**2 + math.fabs(70-y)**2), 1.0])
                pix[x, y] = np.uint8(intensity[0]*255)
                      
        im = Image.fromarray(pix, "L")      
        chromo.fitness = 1.0 / (1.0 + rmsdiff(im, correct_im))
        

def rmsdiff(im1, im2):
    "Calculate the root-mean-square difference between two images"
    diff = ImageChops.difference(im1, im2)
    h = diff.histogram()
    sq = (value*(idx**2) for idx, value in enumerate(h))
    sum_of_squares = sum(sq)
    rms = math.sqrt(sum_of_squares/float(im1.size[0] * im1.size[1]))
    return rms

population.Population.evaluate = eval_fitness

pop = population.Population()
pop.epoch(200, report=True, save_best=False)

winner = pop.stats[0][-1]
print 'Number of evaluations: %d' %winner.id
print "fitness: %1.5f" %winner.fitness


#for ng in winner.node_genes:
 #   print "**************************************************************"
  #  pprint(inspect.getmembers(ng))
    
#conn_list = [(cg.innodeid, cg.outnodeid, cg.weight) for cg in winner.conn_genes if cg.enabled]
 
#print conn_list

visualize.draw_net(winner)


brain = nn.create_ffphenotype(winner)
im = Image.new("L", (4, 4))
pix = im.load()

im_range = np.linspace(-1.0, 1.0, num=4)
im_range2 = np.linspace(1.0, -1.0, num=4)
for xi, x in enumerate(im_range):
    for yi, y in enumerate(im_range2):
        output = brain.sactivate([x,y])
        im.putpixel((xi,yi),output[0]*255)

im = im.resize((512, 512))        
im.save("Test.bmp", "BMP")
im.show()





# Visualize the winner network (requires PyDot)
visualize.draw_net(winner) # best chromosome

# Plots the evolution of the best/average fitness (requires Biggles)
#visualize.plot_stats(pop.stats)
# Visualizes speciation
#visualize.plot_species(pop.species_log)

# Let's check if it's really solved the problem
print '\nBest network output:'
error = 0.0

for i, inputs in enumerate(INPUTS):
    output = brain.sactivate(inputs) # serial activation
    error += (output[0] - OUTPUTS[i])**2
    print "%1.5f \t %1.5f" %(OUTPUTS[i], output[0])
    
print "fitness: %1.5f" %(1/(1+error))

# saves the winner
#file = open('winner_chromosome', 'w')
#pickle.dump(winner, file)
#file.close()
