import math
import cPickle as pickle
from neat import config, population, chromosome, genome, visualize
from neat.nn import nn_pure as nn
import numpy as np
import fourinarow
import os
import datetime
import shutil
import matplotlib.pyplot as plt
from matplotlib import gridspec as gs
from matplotlib import patches as pt
from PIL import Image, ImageChops
#from neat.nn import nn_cpp as nn # C++ extension

config.load('HyperNEAT_test')

#create folder for experiment
now = datetime.datetime.now()
path = os.getcwd() + "\\results\\" + now.strftime("%Y-%m-%d %H%M_%S")
os.makedirs(path)
shutil.copyfile('xor2_config', path +'\\xor2_config')

# set node gene type
chromosome.node_gene_type = genome.NodeGene

# set basic geometry
SUBSTRATE = nn.FeedForward([16,16,1], False, math.tanh)

#coordinates of input layer
INPUTS = [[-1.0, 1.0], [-0.33, 1.0], [0.33, 1.0], [1.0, 1],  
          [-1.0, 0.33], [-0.33, 0.33], [0.33, 0.33], [1.0, 0.33],  
          [-1.0, -0.33], [-0.33, -0.33], [0.33, -0.33], [1.0, -0.33],
          [-1.0, -1.0], [-0.33, -1.0], [0.33, -1.0], [1.0, -1.0]]

#coordinates of hidden layer
HIDDEN =  [[-1.0, 1.0], [-0.33, 1.0], [0.33, 1.0], [1.0, 1],  
          [-1.0, 0.33], [-0.33, 0.33], [0.33, 0.33], [1.0, 0.33],  
          [-1.0, -0.33], [-0.33, -0.33], [0.33, -0.33], [1.0, -0.33],
          [-1.0, -1.0], [-0.33, -1.0], [0.33, -1.0], [1.0, -1.0]]


# produce hyper-cube inputs (coordinates x1 y1 x2 y2)
def fully_connected_coord_input(hypercube_dimensions = 4):
    hc_range = range(hypercube_dimensions)
    return [[x1, y1, x2, y2] for x1 in hc_range for y1 in hc_range for x2 in hc_range for y2 in hc_range]


def eval_fitness(population):
    
    weights_input_to_hidden = np.zeros((16, 16))
    weights_hidden_to_output = np.zeros((16, 1))
    
    for chromo in population:
        net = nn.create_ffphenotype(chromo)

        for i, inputs_i in enumerate(INPUTS):
            for j, inputs_h in enumerate(HIDDEN):
                #net.flush() # not strictly necessary in feedforward nets
                weights_input_to_hidden[i][j] = net.sactivate(inputs_i + inputs_h)[0] # serial activation
                
        for i, inputs in enumerate(HIDDEN):
            weights_hidden_to_output[i] = net.sactivate(inputs + [0,0])[1]
            
        weights = np.append(weights_input_to_hidden.flatten(), weights_hidden_to_output)
        SUBSTRATE.update_weights(list(weights))
           
        fitness = fourinarow.runGame(False, SUBSTRATE, False)
        
        temp = 0.0
        if fitness >= 0.5:
            for i in range(10):
                my_turn = fourinarow.runGame(False, SUBSTRATE, False)
                temp += (my_turn)
       
            fitness = temp / 10.0
            
            print fitness
        
        chromo.fitness = fitness
 

#specify evaluation method
population.Population.evaluate = eval_fitness

#run experiment
pop = population.Population()
pop.epoch(300, report=True, save_best=True, checkpoint_interval=None, save_path=path)

#print stats from winner
winner = pop.stats[0][-1]
print 'Number of evaluations: %d' %winner.id
print "fitness: %1.5f" %winner.fitness

'''
for ng in winner.node_genes:
    print "**************************************************************"
    pprint(inspect.getmembers(ng))
    
conn_list = [(cg.innodeid, cg.outnodeid, cg.weight) for cg in winner.conn_genes if cg.enabled]
 
print conn_list
'''


net = nn.create_ffphenotype(winner)

weights_input_to_hidden = np.zeros((16, 16))
weights_hidden_to_output = np.zeros((16, 1))
pattern_input_to_hidden = np.zeros(16)
pattern_hidden_to_output = np.zeros(16)

for i, inputs_i in enumerate(INPUTS):
    for j, inputs_h in enumerate(HIDDEN):
        net.flush() # not strictly necessary in feedforward nets
        w_ij = net.sactivate(inputs_i + inputs_h)[0]
        pattern_input_to_hidden[j] += w_ij
        weights_input_to_hidden[i][j] = w_ij # serial activation
        
for i, inputs in enumerate(HIDDEN):
    w_i = net.sactivate(inputs + [0,0])[1]
    pattern_hidden_to_output[i] += w_i
    weights_hidden_to_output[i] = w_i
    
weights = np.append(weights_input_to_hidden.flatten(), weights_hidden_to_output)
SUBSTRATE.update_weights(list(weights))

def rolling(a, window):
    shape = (a.size - window + 1, window)
    strides = (a.itemsize, a.itemsize)
    return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)


def draw_board(board):
    gs1 = gs.GridSpec(4,4) 
    gs1.update(wspace=0, hspace=0)
    for x in range(4):
        for y in range(4):
            color = board[x,y]
            ax = plt.subplot(gs1[x,y])
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_ylim(-1,1)
            ax.set_xlim(-1,1)
            if color == 1:
                circle = pt.Circle((0,0), color='red')
                circle.set_radius(0.7)
                ax.add_patch(circle)
            elif color == -1:
                circle = pt.Circle((0,0), color='black')
                circle.set_radius(0.7)
                ax.add_patch(circle)

    plt.show() 

def convert_value(val, dataset_min, dataset_range, output_min, output_range):
    return (((val - dataset_min) * output_range) / dataset_range) + output_min

def plot_influence(a):
    from mpl_toolkits.mplot3d import axes3d
    import matplotlib.pyplot as plt
    from matplotlib.mlab import griddata
    from matplotlib import cm as cm
    from matplotlib import gridspec as gs
    from matplotlib import patches as pt
    import matplotlib
    
    min_a = a.min()
    max_a = a.max()
    range_a = max_a - min_a
    a = [convert_value(val, min_a, range_a, -1, 2) for val in a]
    np.set_printoptions(threshold=np.nan, precision=5, suppress=True)
    from matplotlib import rc
    plt.rcParams['ps.useafm'] = True
    rc('font',**{'family':'sans-serif','sans-serif':['FreeSans']})
    plt.rcParams['pdf.fonttype'] = 42
    

    gs1 = gs.GridSpec(1,3)
    gs1.update(wspace=0.5)

    fig = plt.figure(figsize=(15,10))
    ax = plt.subplot(gs1[0,0:2],projection='3d')
    xi = yi = np.linspace(-1, 1, num=20)
    X, Y = np.meshgrid(xi, yi)
    x = y = np.linspace(-1, 1, num=4)
    ticks = [-1.0, -0.33, 0.33, 1.0]
    Xx, Yy = np.meshgrid(x, y)
    Z = np.rot90(griddata(Xx.flatten(), Yy.flatten(), a, xi, yi))


    ax.plot_surface(X, Y, Z, rstride=1, cstride=1, alpha=1, linewidth=1, cmap=cm.jet)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    
    min = np.min(Z)
    max = np.max(Z)
    zrange = max - min
    
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    ax.set_zlabel('Weight')
    ax.set_zlim(min - (zrange/2),max + (zrange/2) )
    
    
    
    ax = plt.subplot(gs1[0,2], aspect='equal')
    im = ax.imshow(Z, interpolation='nearest', origin='lower', cmap=cm.jet, extent=(-1,1,-1,1))
    CS = ax.contour(X,Y,Z, 6, colors='k')
    ax.clabel(CS, inline=1, fontsize=12, format='%f.2')
    CBI = plt.colorbar(im, orientation='horizontal', shrink=0.8)
    CL = plt.getp(CBI.ax, 'xmajorticklabels')
    plt.setp(CL, fontsize=8)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    
    plt.show()


plot_influence(pattern_hidden_to_output)
plot_influence(pattern_input_to_hidden)

fitness = 0.0    
w = 0
l = 0
t = 0

for i in range(100):
    test = fourinarow.runGame(False, SUBSTRATE, True)
    if test == 1.0: w += 1
    elif test == 0.0: l+= 1
    else: t+=1
        
    fitness += test
    print test
print fitness/100.0

print "Win: {0}  Loss: {1}  Tie: {2}".format(w, l, t)


# Visualize the winner network (requires PyDot)
visualize.draw_net(winner) # best chromosome

# Plots the evolution of the best/average fitness (requires Biggles)
#visualize.plot_stats(pop.stats)
# Visualizes speciation
#visualize.plot_species(pop.species_log)


# saves the winner
file = open(path +'\\winner_chromosome ' + str(winner.fitness), 'w')
pickle.dump(winner, file)
file.close()

if __name__ == '__main__':
    print 'main'
