# Four-In-A-Row (a Connect Four clone)
# By Al Sweigart al@inventwithpython.com
# http://inventwithpython.com/pygame
# Released under a "Simplified BSD" license

#Original AI versus HyperNEAT (AI versus AI)
#Modified:   08.02.2012 
#By:         Stian P. Kvaale
#E-mail:     stianped@stud.ntnu.no

import random, copy
import numpy as np
from mpl_toolkits.mplot3d import axes3d
import matplotlib.pyplot as plt
from matplotlib.mlab import griddata
from matplotlib import cm as cm
from matplotlib import gridspec as gs
from matplotlib import patches as pt
import matplotlib

INPUTS = [[-1.0, 1.0], [-0.33, 1.0], [0.33, 1.0], [1.0, 1],  
          [-1.0, 0.33], [-0.33, 0.33], [0.33, 0.33], [1.0, 0.33],  
          [-1.0, -0.33], [-0.33, -0.33], [0.33, -0.33], [1.0, -0.33],
          [-1.0, -1.0], [-0.33, -1.0], [0.33, -1.0], [1.0, -1.0]]

#coordinates of hidden layer
HIDDEN =  [[-1.0, 1.0], [-0.33, 1.0], [0.33, 1.0], [1.0, 1],  
          [-1.0, 0.33], [-0.33, 0.33], [0.33, 0.33], [1.0, 0.33],  
          [-1.0, -0.33], [-0.33, -0.33], [0.33, -0.33], [1.0, -0.33],
          [-1.0, -1.0], [-0.33, -1.0], [0.33, -1.0], [1.0, -1.0]]

BOARDWIDTH = 4  # how many spaces wide the board is
BOARDHEIGHT = 4 # how many spaces tall the board is
assert BOARDWIDTH >= 4 and BOARDHEIGHT >= 4, 'Board must be at least 4x4.'

DIFFICULTY = 3 # how many moves to look ahead. (>2 is usually too much)

RED = 1
BLACK = -1
EMPTY = 0
HUMAN = 'human'
COMPUTER = 'computer'


def main(ANN):
    runGame(False, ANN)



def runGame(isFirstGame, ANN, document_winner):
    if isFirstGame:
        # Let the computer go first on the first game, so the player
        # can see how the tokens are dragged from the token piles.
        turn = COMPUTER
        showHelp = True
    else:
        # Randomly choose who goes first.
        turn = HUMAN
        showHelp = False

    # Set up a blank board data structure.
    mainBoard = getNewBoard()
    first =True
    history= []
    #ANN.sactivate(mainBoard.flatten())

    while True: # main game loop
        #print mainBoard
        
        if turn == HUMAN:
            # Human player's turn.
            '''if first is not True and document_winner == True:
                board_pre = copy.deepcopy(mainBoard)
                ANN.sactivate(board_pre.flatten())
                pattern = get_pattern(ANN)
                column = get_ANN_Move(mainBoard, showHelp, ANN)
                makeMove(mainBoard, RED, column)
                plot_influence(board_pre, mainBoard, pattern)
                
            else:
                first = False
                column = get_ANN_Move(mainBoard, showHelp, ANN)
                makeMove(mainBoard, RED, column)'''
            
            pre_state = np.copy(mainBoard)
            column = get_ANN_Move(mainBoard, showHelp, ANN)
            makeMove(mainBoard, RED, column)
            history.append((pre_state, np.copy(mainBoard)))
            
            if showHelp:
                # turn off help arrow after the first move
                showHelp = False
            if isWinner(mainBoard, RED):
                #winning and returning perfect fitness
                #print mainBoard
                if document_winner : replay_and_plot(ANN, history)
                return 1.0
            turn = COMPUTER # switch to other player's turn

        else:
            # Computer player's turn.
            column = getComputerMove(mainBoard)
            makeMove(mainBoard, BLACK, column)
            if isWinner(mainBoard, BLACK):
                #losing and returning negative fitness + number of n pieces in a row
                n_in_a_row = count_n_in_a_row(mainBoard, RED)
                return 0.0
            turn = HUMAN # switch to other player's turn
            

        if isBoardFull(mainBoard):
            # A completely filled board means it's a tie.
            #Returning 0.5 fitness + number of n pieces in a row
            n_in_a_row = count_n_in_a_row(mainBoard, RED) 
            return 0.5 #+ (n_in_a_row[0] / 1000.0) + (n_in_a_row[1] / 100.0) 
        
        
def count_n_in_a_row(board, color):
    return np.add(np.add(horizontal_search(board, color), vertical_search(board, color)), diagonal_search(board, color))
    
    
def horizontal_search(board, color):
    n_in_a_row= [0] * 4
    for row in range(BOARDWIDTH):
        n = 0
        for column in range(BOARDHEIGHT):
            if board[row, column] == color:
                n += 1
            else:
                n= 0
        if n > 1:
            n_in_a_row[n] += 1
        
    return n_in_a_row

def vertical_search(board, color):
    n_in_a_row= [0] * 4
    for column in range(BOARDHEIGHT):
        n = 0
        for row in range(BOARDWIDTH):
            if board[row, column] == color:
                n += 1
            else:
                n = 0
        if n > 1:
            n_in_a_row[n] += 1
        
    return n_in_a_row

def diagonal_search(board, color):
    dupeBoard = copy.deepcopy(board)
    return np.add(count_diagonal(dupeBoard, color), count_diagonal(dupeBoard[::-1], color))   
            
def makeMove(board, player, column):
    lowest = getLowestEmptySpace(board, column)
    if lowest != -1:
        board[column][lowest] = player
  

def getNewBoard():
    board = []
    for x in range(BOARDWIDTH):
        board.append([EMPTY] * BOARDHEIGHT)
    return np.array(board)


def get_ANN_Move(board, isFirstMove, ANN):
    potentialMoves = getPotentialMoves_ANN(board, RED, DIFFICULTY, ANN)
    return potentialMoves.index(max(potentialMoves))

def getComputerMove(board):
    potentialMoves = getPotentialMoves(board, BLACK, DIFFICULTY)
    # get the best fitness from the potential moves
    bestMoveFitness = -5
    for i in range(BOARDWIDTH):
        if potentialMoves[i] > bestMoveFitness and isValidMove(board, i):
            bestMoveFitness = potentialMoves[i]
    # find all potential moves that have this best fitness
    bestMoves = []
    for i in range(len(potentialMoves)):
        if potentialMoves[i] == bestMoveFitness and isValidMove(board, i):
            bestMoves.append(i)
    return random.choice(bestMoves)


def getPotentialMoves(board, tile, lookAhead):
    if lookAhead == 0 or isBoardFull(board):
        return [0] * BOARDWIDTH

    if tile == RED:
        enemyTile = BLACK
    else:
        enemyTile = RED

    # Figure out the best move to make.
    potentialMoves = [0] * BOARDWIDTH
    for firstMove in range(BOARDWIDTH):
        dupeBoard = copy.deepcopy(board)
        if not isValidMove(dupeBoard, firstMove):
            continue
        makeMove(dupeBoard, tile, firstMove)
        if isWinner(dupeBoard, tile):
            # a winning move automatically gets a perfect fitness
            potentialMoves[firstMove] = 1
            break # don't bother calculating other moves
        else:
            # do other player's counter moves and determine best one
            if isBoardFull(dupeBoard):
                potentialMoves[firstMove] = 0
            else:
                for counterMove in range(BOARDWIDTH):
                    dupeBoard2 = copy.deepcopy(dupeBoard)
                    if not isValidMove(dupeBoard2, counterMove):
                        continue
                    makeMove(dupeBoard2, enemyTile, counterMove)
                    if isWinner(dupeBoard2, enemyTile):
                        # a losing move automatically gets the worst fitness
                        potentialMoves[firstMove] = -1
                        break
                    else:
                        # do the recursive call to getPotentialMoves()
                        results = getPotentialMoves(dupeBoard2, tile, lookAhead - 1)
                        potentialMoves[firstMove] += (sum(results) / BOARDWIDTH) / BOARDWIDTH
    return potentialMoves

def getPotentialMoves_ANN(board, tile, lookAhead, ANN):
    potentialMoves = [0] * BOARDWIDTH
    for move in range(BOARDWIDTH):
        dupeBoard = copy.deepcopy(board)
        makeMove(dupeBoard, RED, move)
        if isValidMove(dupeBoard, move):
            
            #print dupeBoard
            
            potentialMoves[move] = ANN.sactivate(dupeBoard.flatten())
            
    return potentialMoves     

def getLowestEmptySpace(board, column):
    # Return the row number of the lowest empty row in the given column.
    for y in range(BOARDHEIGHT-1, -1, -1):
        if board[column][y] == EMPTY:
            return y
    return -1


def isValidMove(board, column):
    # Returns True if there is an empty space in the given column.
    # Otherwise returns False.
    if column < 0 or column >= (BOARDWIDTH) or board[column][0] != EMPTY:
        return False
    return True


def isBoardFull(board):
    # Returns True if there are no empty spaces anywhere on the board.
    for x in range(BOARDWIDTH):
        for y in range(BOARDHEIGHT):
            if board[x][y] == EMPTY:
                return False
    return True


def isWinner(board, tile):
    # check horizontal spaces
    for x in range(BOARDWIDTH - 3):
        for y in range(BOARDHEIGHT):
            if board[x][y] == tile and board[x+1][y] == tile and board[x+2][y] == tile and board[x+3][y] == tile:
                return True
    # check vertical spaces
    for x in range(BOARDWIDTH):
        for y in range(BOARDHEIGHT - 3):
            if board[x][y] == tile and board[x][y+1] == tile and board[x][y+2] == tile and board[x][y+3] == tile:
                return True
    # check / diagonal spaces
    for x in range(BOARDWIDTH - 3):
        for y in range(3, BOARDHEIGHT):
            if board[x][y] == tile and board[x+1][y-1] == tile and board[x+2][y-2] == tile and board[x+3][y-3] == tile:
                return True
    # check \ diagonal spaces
    for x in range(BOARDWIDTH - 3):
        for y in range(BOARDHEIGHT - 3):
            if board[x][y] == tile and board[x+1][y+1] == tile and board[x+2][y+2] == tile and board[x+3][y+3] == tile:
                return True
    return False


def count_diagonal(board, color):
    
    num_in_a_row = [0] * 4
    
    for i in range(-4, 5, 1):
        count = list(np.diagonal(board, i)).count(color)
        if count is not 0 and count is not 1:
             num_in_a_row[count] += 1
        
    return num_in_a_row

      

def draw_board(board):
    gs1 = gs.GridSpec(4,4) 
    gs1.update(wspace=0, hspace=0)
    
    ax = plt.subplot(gs1[x,y])
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_ylim(-1,1)
    ax.set_xlim(-1,1)
    ax.grid(True)
    for x in range(4):
        for y in range(4):
            color = board[x,y]
            
            if color == 1:
                circle = pt.Circle((0,0), color='red')
                circle.set_radius(0.1)
                ax.add_patch(circle)
            elif color == -1:
                circle = pt.Circle((0,0), color='black')
                circle.set_radius(0.1)
                ax.add_patch(circle)
                
    

    plt.show()  
    
def get_pattern(ANN):
    return ANN.get_receptive_fields(16,16), ANN.get_hidden_activation_levels(16,16)         
        
def plot_influence(board_pre, board_post, patterns1, patterns2):
    np.set_printoptions(threshold=np.nan, precision=5, suppress=True)
    from matplotlib import rc
    plt.rcParams['ps.useafm'] = True
    rc('font',**{'family':'sans-serif','sans-serif':['FreeSans']})
    plt.rcParams['pdf.fonttype'] = 42
    (a,b), c = patterns1
    
    gs1 = gs.GridSpec(2,3,top=0.95, bottom=0.35, hspace=0.05)
    gs1.update(wspace=0.25, hspace=0.30)
    fig = plt.figure(figsize=(15,10))
    
    xi = yi = np.linspace(-1, 1, num=50)
    X, Y = np.meshgrid(xi, yi)
    x = y = np.linspace(-1, 1, num=4)
    ticks = [-1.0, -0.33, 0.33, 1.0]
    Xx, Yy = np.meshgrid(x, y)
    Z = np.rot90(griddata(Xx.flatten(), Yy.flatten(), a, xi, yi))
    ax = plt.subplot(gs1[0,0], aspect='equal')
    im = ax.imshow(Z, interpolation='nearest', origin='lower', cmap=cm.jet, extent=(-1,1,-1,1))
    CS = ax.contour(X,Y,Z, 6, colors='k')
    ax.clabel(CS, inline=1, fontsize=12, format='%f.2')
    CBI = plt.colorbar(im, orientation='vertical', shrink=0.64)
    CL = plt.getp(CBI.ax, 'xmajorticklabels')
    plt.setp(CL, fontsize=7)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    ax.set_title('(a) Input-to-Hidden')
    ax.grid(True)
    
    xi = yi = np.linspace(-1, 1, num=50)
    X, Y = np.meshgrid(xi, yi)
    x = y = np.linspace(-1, 1, num=4)
    ticks = [-1.0, -0.33, 0.33, 1.0]
    Xx, Yy = np.meshgrid(x, y)
    Z = np.rot90(griddata(Xx.flatten(), Yy.flatten(), b, xi, yi))
    ax = plt.subplot(gs1[0,1], aspect='equal')
    im = ax.imshow(Z, interpolation='nearest', origin='lower', cmap=cm.jet, extent=(-1,1,-1,1))
    CS = ax.contour(X,Y,Z, 6, colors='k')
    ax.clabel(CS, inline=1, fontsize=12, format='%f.2')
    CBI = plt.colorbar(im, orientation='vertical', shrink=0.64)
    CL = plt.getp(CBI.ax, 'xmajorticklabels')
    plt.setp(CL, fontsize=7)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    ax.set_title('(b) Hidden-to-Output')
    ax.grid(True)
    
    xi = yi = np.linspace(-1, 1, num=50)
    X, Y = np.meshgrid(xi, yi)
    x = y = np.linspace(-1, 1, num=4)
    ticks = [-1.0, -0.33, 0.33, 1.0]
    Xx, Yy = np.meshgrid(x, y)
    Z = np.rot90(griddata(Xx.flatten(), Yy.flatten(), c, xi, yi))
    ax = plt.subplot(gs1[0,2], aspect='equal')
    im = ax.imshow(Z, interpolation='nearest', origin='lower', cmap=cm.jet, extent=(-1,1,-1,1))
    CS = ax.contour(X,Y,Z, 6, colors='k')
    ax.clabel(CS, inline=1, fontsize=12, format='%f.2')
    CBI = plt.colorbar(im, orientation='vertical', shrink=0.64)
    CL = plt.getp(CBI.ax, 'xmajorticklabels')
    plt.setp(CL, fontsize=7)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    ax.set_title('(c) Hidden layer Activation levels')
    ax.grid(True)
    
    (a,b), c = patterns2
    xi = yi = np.linspace(-1, 1, num=50)
    X, Y = np.meshgrid(xi, yi)
    x = y = np.linspace(-1, 1, num=4)
    ticks = [-1.0, -0.33, 0.33, 1.0]
    Xx, Yy = np.meshgrid(x, y)
    Z = np.rot90(griddata(Xx.flatten(), Yy.flatten(), a, xi, yi))
    ax = plt.subplot(gs1[1,0], aspect='equal')
    im = ax.imshow(Z, interpolation='nearest', origin='lower', cmap=cm.jet, extent=(-1,1,-1,1))
    CS = ax.contour(X,Y,Z, 6, colors='k')
    ax.clabel(CS, inline=1, fontsize=12, format='%f.2')
    CBI = plt.colorbar(im, orientation='vertical', shrink=0.64)
    CL = plt.getp(CBI.ax, 'xmajorticklabels')
    plt.setp(CL, fontsize=7)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    ax.set_title('(d) Input-to-Hidden')
    ax.grid(True)
    
    xi = yi = np.linspace(-1, 1, num=50)
    X, Y = np.meshgrid(xi, yi)
    x = y = np.linspace(-1, 1, num=4)
    ticks = [-1.0, -0.33, 0.33, 1.0]
    Xx, Yy = np.meshgrid(x, y)
    Z = np.rot90(griddata(Xx.flatten(), Yy.flatten(), b, xi, yi))
    ax = plt.subplot(gs1[1,1], aspect='equal')
    im = ax.imshow(Z, interpolation='nearest', origin='lower', cmap=cm.jet, extent=(-1,1,-1,1))
    CS = ax.contour(X,Y,Z, 6, colors='k')
    ax.clabel(CS, inline=1, fontsize=12, format='%f.2')
    CBI = plt.colorbar(im, orientation='vertical', shrink=0.64)
    CL = plt.getp(CBI.ax, 'xmajorticklabels')
    plt.setp(CL, fontsize=7)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    ax.set_title('(e) Hidden-to-Output')
    ax.grid(True)
    
    xi = yi = np.linspace(-1, 1, num=50)
    X, Y = np.meshgrid(xi, yi)
    x = y = np.linspace(-1, 1, num=4)
    ticks = [-1.0, -0.33, 0.33, 1.0]
    Xx, Yy = np.meshgrid(x, y)
    Z = np.rot90(griddata(Xx.flatten(), Yy.flatten(), c, xi, yi))
    ax = plt.subplot(gs1[1,2], aspect='equal')
    im = ax.imshow(Z, interpolation='nearest', origin='lower', cmap=cm.jet, extent=(-1,1,-1,1))
    CS = ax.contour(X,Y,Z, 6, colors='k')
    ax.clabel(CS, inline=1, fontsize=12, format='%f.2')
    CBI = plt.colorbar(im, orientation='vertical', shrink=0.64)
    CL = plt.getp(CBI.ax, 'xmajorticklabels')
    plt.setp(CL, fontsize=7)
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xlabel('X')
    ax.set_xlim(-1, 1)
    ax.set_ylabel('Y')
    ax.set_ylim(-1, 1)
    ax.set_title('(f) Hidden layer Activation levels')
    ax.grid(True)
    
    gs1 = gs.GridSpec(1,2, top=0.27, bottom=0.02)
    ticks=[1,2,3,4]
    
    ax = plt.subplot(gs1[0,0], aspect='equal')
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    
    for label in ax.get_xticklabels() + ax.get_yticklabels(): 
        label.set_fontsize(0)

    ax.set_ylim(4,0)
    ax.set_xlim(4,0)
    ax.set_title('(g) Pre-move board state')
    ax.grid(True, color='black', linestyle='-', linewidth=1)
    for x in range(3,-1,-1):
        for y in range(3,-1,-1):
            color = board_pre[x,y]
            
            if color == 1:
                circle = pt.Circle((x+0.5,y+0.5), color='red')
                circle.set_radius(0.3)
                ax.add_patch(circle)
            elif color == -1:
                circle = pt.Circle((x+0.5,y+0.5), color='black')
                circle.set_radius(0.3)
                ax.add_patch(circle)
                
    ax = plt.subplot(gs1[0,1], aspect='equal')
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    for label in ax.get_xticklabels() + ax.get_yticklabels(): 
        label.set_fontsize(0)
    ax.set_ylim(4,0)
    ax.set_xlim(4,0)
    ax.set_title('(h) Post-move board state')
    ax.grid(True, color='black', linestyle='-', linewidth=1)
    for x in range(3,-1,-1):
        for y in range(3,-1,-1):
            color = board_post[x,y]
            
            if color == 1:
                circle = pt.Circle((x+0.5,y+0.5), color='red')
                circle.set_radius(0.3)
                ax.add_patch(circle)
            elif color == -1:
                circle = pt.Circle((x+0.5,y+0.5), color='black')
                circle.set_radius(0.3)
                ax.add_patch(circle)
    
    plt.show()
    


def replay_and_plot(ANN, board_states):
    for pre_state, post_state in board_states[1:]:
        '''
        potentialMoves = [0] * BOARDWIDTH
        for move in range(BOARDWIDTH):
            dupeBoard = copy.deepcopy(pre_state)
            makeMove(dupeBoard, RED, move)
            if isValidMove(dupeBoard, move):

                potentialMoves[move] = ANN.sactivate(dupeBoard.flatten())
                pattern = get_pattern(ANN)   
                plot_influence(pre_state, dupeBoard, pattern)
 
        '''
        ANN.sactivate(pre_state.flatten())
        pattern1 = get_pattern(ANN)   

        ANN.sactivate(post_state.flatten())

        pattern2 = get_pattern(ANN)   

        plot_influence(pre_state, post_state, pattern1, pattern2)
        
            
            
    
    

                  

if __name__ == '__main__':
    ''''a = numpy.array([['red','black','black', 'black'],
                     ['black', 'red','black', 'black'], 
                     ['black', 'red', 'red', 'red'],
         ['black', 'black', 'red', 'black']])
    
   
    print numpy.add(count_diagonal(a, 'red'), count_diagonal(a[::-1], 'red'))
    
    print a.flatten()
    
    b = numpy.array([[ 0,  0,  0, -1],
                     [ 1, -1, -1,  1],
                     [ 0,  1, -1, -1],
                     [ 0, -1,  1,  1],
                     [ 1, -1,  1,  1]])

    
    print horizontal_search(b, RED)
    print vertical_search(b, RED)
    
    print numpy.add(numpy.add(horizontal_search(b, RED), vertical_search(b, RED)), diagonal_search(b, RED))
    
    print diagonal_search(b, RED)''' 
    
    b = np.array([[ 0,  0,  0, -1],
                     [ 1, -1, -1,  1],
                     [ 0,  1, -1, -1],
                     [ 0, -1,  1,  1],
                     [ 1, -1,  1,  1]])        
       
    draw_board(b)
        
        
    
    #main()
